/*
 * Copyright 1996-2002 netFluid Technology Ltd
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.1.1.1 $  $Date: 2003/01/10 23:16:04 $
 *
 * Author: David Zambonini
 *
 * ---------------------------------------------------------------------------
 */

#include <phlib/memdbinc.h>

/* Allocation tag - verifies we have a debug allocation... set to anything
 *  untypical 
 */
#define MEM_DEBUG_TAG 0xC10DF9BEUL

static mem_db_info *first = NULL; /* Storage for debug chain pointer */

/* Maximum allowable size passed via size_t - don't worry, the result will
 * be correct, even if the compiler complains 
 */
const unsigned long max_alloc=((1UL<<(8*sizeof(size_t)))-sizeof(mem_db_info));

static unsigned long big_alloc = 0; /* Biggest allocation made */
static unsigned long n_active  = 0; /* Number of active blocks */
static unsigned long n_alloc   = 0; /* Number of successful allocations */
static unsigned long n_dealloc = 0; /* Number of deallocations */
static unsigned long n_failure = 0; /* Number of failed allocations */

static int registered    = 0;       /* Post-debug procedure registered flag */
static int serious_error = 0;       /* Serious memory error flag            */

/* Post-debug procedure prototype 
 */
static void mem_db_results(void);

/* ------------------------------------------------------------------------- */
/* Ensure debugging active 
 */
void 
set_mem_db_used(void)
{
  if (registered == 0 ) {
    if( atexit( mem_db_results ) != 0 ) {
      fprintf(stderr,"Could not register memory debug exit function\n");
      exit(EXIT_FAILURE);
    }
    registered = 1;
  }
}

/* ------------------------------------------------------------------------- */
/* Memory allocation failure 
 */
void 
set_mem_db_failure(void)
{
  n_failure++;
}

/* ------------------------------------------------------------------------- */
/* Create a new memory information node - used by malloc,calloc 
 */
void 
set_mem_db_info(mem_db_info  *ptr,
		const char   *allocator,
		size_t        size,
		const char   *filename,
		unsigned long line)
{
  /* Set node debug info */
  ptr->id_tag=MEM_DEBUG_TAG;
  ptr->line=line;
  ptr->filename=filename;
  ptr->allocator=allocator;
  ptr->size=size;

    /* Link into chain */
  if( first != NULL ) {
    /* Just another node */
    ptr->last=first;
    ptr->next=first->next;
    first->next->last=ptr;
    first->next=ptr;
  } else {
    /* First node */
    ptr->last=ptr;
    ptr->next=ptr;
    first=ptr;
  }

  /* Update allocation stats */
  if( size > big_alloc ) {
    big_alloc=size;
  }
  n_alloc++;
  n_active++;
}

/* ------------------------------------------------------------------------- */
/* Move an existing memory information node - used by realloc 
 */
void 
move_mem_db_info(mem_db_info *from,
		 mem_db_info *ptr,
		 const char *allocator,
		 size_t size,
		 const char *filename,
		 unsigned long line)
{
  /* Update links */
  if( ptr->next != from ) {
    ptr->next->last=ptr;
    ptr->last->next=ptr;
  } else {
    /* Only one entry, link up to itself */
    ptr->next=ptr->last=ptr;
  }

  /* Update internal info */
  ptr->line=line;
  ptr->filename=filename;
  ptr->allocator=allocator;
  ptr->size=size;

    /* Check first node link */
  if( first == from ) {
    first=ptr;
  }

  /* Update allocation stats */
  if( size > big_alloc ) {
    big_alloc=size;
  }
}

/* ------------------------------------------------------------------------- */
/* Check an existing node ID 
 */
void 
check_mem_db_info(mem_db_info *ptr,const char *filename,unsigned long line)
{
  if( ptr->id_tag != MEM_DEBUG_TAG ) {
    /* How did this get in here? We might be mixing code, or more
       seriously, the heap could be corrupt */
    fprintf(stderr,"A debug information tag check failed at %s, line %lu\n"
	    "Either the heap has been corrupted, an address freed "
	    "more than once,\nor an invalid "
	    "address was given\n",filename,line);
    serious_error=1;
    exit(EXIT_FAILURE);
  }
}

/* ------------------------------------------------------------------------- */
/* Remove an existing memory information node - used by free 
 */
void 
unset_mem_db_info(mem_db_info *ptr)
{
  /* Unset id tag */
  ptr->id_tag=0;
    
    /* Unlink from chain */
  ptr->next->last=ptr->last;
  ptr->last->next=ptr->next;

  if( ptr == first ) {
    /* This is the first node */
    if( ptr->next == ptr ) {
      /* It's also the only node */
      first=NULL;
      if( n_active != 1 ) {
	/* So why don't our stats match up? */
	fprintf(stderr,"Error detected in memory debug, active memory "
		"information corrupted\n");
	serious_error=1;
	exit(EXIT_FAILURE);
      }
    } else {
      /* Set first to the next available node */
      first=ptr->next;
    }
  }
  /* Update debug info */
  n_active--;
  n_dealloc++;
}

/* ------------------------------------------------------------------------- */
/* Exit routine reports any problems found in program 
 */
static void 
mem_db_results(void)
{
  /* Print stats. */
  fprintf(stderr,"\nMemory allocation debug statistics:\n"
	  "\tMaximum allowable block size               : %lu bytes\n"
	  "\tLargest allocation block created           : %lu bytes\n"
	  "\tNumber of successful allocation requests   : %lu\n"
	  "\tNumber of successful deallocation requests : %lu\n"
	  "\tNumber of failed allocation requests       : %lu\n"
	  "\tNumber of allocation blocks still active   : %lu\n\n"
	  ,max_alloc
	  ,big_alloc
	  ,n_alloc
	  ,n_dealloc
	  ,n_failure
	  ,n_active);

  /* Is there any still-allocated memory? */
  if( (first == NULL) && (n_active == 0) ) {
    /* Nope, no problems */
    fprintf(stderr,"Memory debug reports no problems found\n");

  } else if (serious_error == 0 ) {
    /* Problems found - memory still allocated */
    fprintf(stderr,"The following memory problems were detected:\n");
    if (first == NULL ) {
      /* Ah... a very serious problem */
      fprintf(stderr,"\tAllocated memory blocks exist, but cannot be found!\n");
    } else {
      /* Not so serious, we've merely failed to deallocate */
      unsigned long n_found=0;
      mem_db_info *ptr=first;
	  
      do {
	int n;
	const unsigned char *cptr;
	if( ptr->id_tag != MEM_DEBUG_TAG) {
	  /* Oh dear... wrong! serious problem, info has been corrupted */
	  fprintf(stderr,"\tA corrupted block was found, cannot proceed\n");
	  return;
	}
	/* Print block stats */
	fprintf(stderr,"\tA %i byte block %sed in %s, "
		"line %lu was still active\n",
		ptr->size,
		ptr->allocator,
		ptr->filename,
		ptr->line);
	/* Print block contents */
	fprintf(stderr,"\t\t( Block contains '");
	cptr=((const unsigned char*)ptr)+sizeof(mem_db_info);
	for (n=0; ( (n < 40) && (n < (ptr->size)) ); n++ ) {
	  fprintf(stderr,"%c",((*cptr>0x1F)&&(*cptr<0x80))?*cptr:'?');
	  cptr++;
	}
	fprintf(stderr,"'... )\n");

	/* Next node, please */
	ptr=ptr->next;
	n_found++;
      }
      while (ptr!=first);
      if( n_found != n_active ) {
	/* Ah.. unfortuantely, the chain seems be wrong */
	fprintf(stderr,"\tNumber of information blocks found does not "
		"match recorded total!\n");
      }
    }
  } else {
    /* Serious problem halted operation */
    fprintf(stderr,"A serious error halted operation\n");
  }
}

/* ------------------------------------------------------------------------- */
/* end */
